/****************************************************************************************
*  YAROCK                                                                               *
*  Copyright (c) 2010-2013 Sebastien amardeilh <sebastien.amardeilh+yarock@gmail.com>   *
*                                                                                       *
*  This program is free software; you can redistribute it and/or modify it under        *
*  the terms of the GNU General Public License as published by the Free Software        *
*  Foundation; either version 2 of the License, or (at your option) any later           *
*  version.                                                                             *
*                                                                                       *
*  This program is distributed in the hope that it will be useful, but WITHOUT ANY      *
*  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A      *
*  PARTICULAR PURPOSE. See the GNU General Public License for more details.             *
*                                                                                       *
*  You should have received a copy of the GNU General Public License along with         *
*  this program.  If not, see <http://www.gnu.org/licenses/>.                           *
*****************************************************************************************/
#include "playlistmodel.h"
#include "playlistproxymodel.h"
#include "task_manager.h"

#include "media_search_engine.h"  // for sorting

#include "utilities.h"
#include "debug.h"


/*******************************************************************************
 SortByRowAsc function object
   -> avoid to use boost to bind member function in qSort
*******************************************************************************/
struct SortByRowAsc
{
    SortByRowAsc(PlaylistModel* model) : m_model(model) { }

    bool operator()(MEDIA::TrackPtr trackleft, MEDIA::TrackPtr trackright)
    {
          return m_model->rowForTrack(trackleft) < m_model->rowForTrack(trackright);
    }
    
private:
    PlaylistModel *m_model;
};


/*
********************************************************************************
*                                                                              *
*    Class PlaylistModel                                                       *
*                                                                              *
********************************************************************************
*/
PlaylistModel::PlaylistModel(QObject* parent) : QAbstractListModel(parent)
{
    m_playing_track      =  MEDIA::TrackPtr(0);
    m_requested_track    =  MEDIA::TrackPtr(0);
    m_stop_after_track   =  MEDIA::TrackPtr(0);

    m_proxy_model = new PlaylistProxyModel(this);
    m_proxy_model->setSourceModel(this);
    
    m_task_manager = new TaskManager(this);
}

QSortFilterProxyModel* PlaylistModel::proxy() const
{
    return m_proxy_model;
}

TaskManager* PlaylistModel::manager() const
{
    return m_task_manager;
}

/*******************************************************************************
    PlaylistModel::clear
*******************************************************************************/
void PlaylistModel::clear()
{
    //Debug::debug() << " --- PlaylistModel->clear";
    removeRows(0, rowCount(QModelIndex()), QModelIndex());

    m_tracks.clear();
    m_playing_track     =  MEDIA::TrackPtr(0);
    m_requested_track   =  MEDIA::TrackPtr(0);

    emit dataChanged(QModelIndex(), QModelIndex());
    emit updated();
}

/*******************************************************************************
    QAbstractListModel implementation
*******************************************************************************/
int PlaylistModel::rowCount(const QModelIndex & /* parent */) const
{
    return m_tracks.size();
}

QVariant PlaylistModel::data(const QModelIndex &index, int role) const
{
Q_UNUSED(index)
Q_UNUSED(role)
    return QVariant();
}

int PlaylistModel::rowForTrack(const MEDIA::TrackPtr track)
{
    return m_tracks.indexOf(track);
}

bool PlaylistModel::removeRows(int position, int count, const QModelIndex & /*parent*/)
{
    //Debug::debug() << "--- PlaylistModel-->removeRow position rows" << position;
    //Debug::debug() << "--- PlaylistModel-->removeRow position count" << count;
    if( position + count - 1 >= m_tracks.size()) return false;

    m_mutex.lock();
    beginRemoveRows(QModelIndex(), position, position + count - 1);
    for (int itemcount = (count-1); itemcount >= 0; itemcount--) {
      MEDIA::TrackPtr mi = m_tracks.takeAt(position + itemcount);
      mi.reset();
    }
    endRemoveRows();
    m_mutex.unlock();

    return true;
}

/*******************************************************************************
    PlaylistModel::updateBrokenItem
*******************************************************************************/
void PlaylistModel::updateBrokenItem(MEDIA::TrackPtr tk)
{
    //Debug::debug() << "PlaylistModel -> updateBrokenItem";
    if(!tk) return;

    foreach(MEDIA::TrackPtr track, m_tracks) {
      if(track == tk) {
        track->isBroken  = true;
        track->isPlaying = false;        

        int row = m_tracks.indexOf(track);
        QModelIndex newIndex = index(row, 0, QModelIndex());
        emit dataChanged(newIndex, newIndex);

        break;
      }
    }
}

/*******************************************************************************
    PlaylistModel::updatePlayingItem
*******************************************************************************/
void PlaylistModel::updatePlayingItem(MEDIA::TrackPtr tk)
{
    //Debug::debug() << "PlaylistModel -> updatePlayingItem";
    stopPlayingItem();
    if(!tk) return;
        
    foreach(MEDIA::TrackPtr track, m_tracks) {
      if(track == tk) {
        track->isBroken  = false;
        track->isPlaying = true;
        track->isPlayed  = true;
              
        m_playing_track = track;
        
        int row = m_tracks.indexOf(track);
        QModelIndex newIndex = index(row, 0, QModelIndex());
        emit dataChanged(newIndex, newIndex);

        break;
      }
    }
}

/*******************************************************************************
    PlaylistModel::stopPlayingItem
*******************************************************************************/
void PlaylistModel::stopPlayingItem()
{
    //! unset previous playing item
    if(m_playing_track)
    {
      m_playing_track->isPlaying = false;
      
      int row = m_tracks.indexOf(m_playing_track);
      QModelIndex newIndex = index(row, 0, QModelIndex());
      emit dataChanged(newIndex, newIndex);

      m_playing_track =  MEDIA::TrackPtr(0);
    }
}


/*******************************************************************************
    PlaylistModel::unplayedMediaItem
*******************************************************************************/
QList<MEDIA::TrackPtr> PlaylistModel::unplayedMediaItem()
{
    Debug::debug() << "--- PlaylistModel::unplayedMediaItem";

    QList<MEDIA::TrackPtr>  list;
    foreach (MEDIA::TrackPtr track, m_tracks) {
      if( track->isPlayed == false && filterContainsRow(rowForTrack(track)))
        list << track;
    }

    return list;
}

/*******************************************************************************
    PlaylistModel::setRequestedMediaItem
*******************************************************************************/
void PlaylistModel::setRequestedMediaItem(int row)
{
    if (!rowExists(row)) 
      m_requested_track = MEDIA::TrackPtr(0);
    else
      m_requested_track = m_tracks.at(row);
}

MEDIA::TrackPtr PlaylistModel::getRequestedMediaItem() const
{
    return m_requested_track;
}


/*******************************************************************************
    PlaylistModel::filterContainsRow
*******************************************************************************/
bool PlaylistModel::filterContainsRow(int source_row)
{
    if (source_row < 0 || source_row >= rowCount())
      return false;
    return m_proxy_model->filterAcceptsRow(source_row, QModelIndex());
}

/*******************************************************************************
    PlaylistModel::skipBackward
*******************************************************************************/
int PlaylistModel::skipBackward(bool repeat /*=false*/)
{
   /* tant que l'on ne trouve pas de ligne située avant presente dans le proxy */
   int i = rowForTrack(m_requested_track); /* source row requested */ 
   --i;
   while(i >= 0 && !filterContainsRow(i))
     --i;
   
   if(repeat && i == -1) {
     int j = rowCount() -1;
     while(j > i && !filterContainsRow(j))
       --j;
       
     i = j;
   }
   return i; 
}

/*******************************************************************************
    PlaylistModel::skipForward
*******************************************************************************/
int PlaylistModel::skipForward(bool repeat /*=false*/)
{
   /* tant que l'on ne trouve pas de ligne située aprés presente dans le proxy */
   int i = rowForTrack(m_requested_track); /* source row requested */ 
   int source_row  = i;
   ++i;
   while(i < rowCount() && !filterContainsRow(i))
     ++i;

   if(i >= rowCount()) i = -1;
   
   if(repeat && i == -1) {
     int j = 0;
     while(j < source_row && !filterContainsRow(j))
       ++j;

     i = j;
   }
   return i;    
}

/*******************************************************************************
    PlaylistModel::MediaItemAt
*******************************************************************************/
MEDIA::TrackPtr PlaylistModel::MediaItemAt(int row) const
{
    if (rowExists(row))
      return m_tracks.at(row);

    return MEDIA::TrackPtr(0);
}

/*******************************************************************************
    PlaylistModel::getPlayingRow
*******************************************************************************/
int PlaylistModel::getPlayingRow()
{
    if(m_playing_track)
      return m_tracks.indexOf(m_playing_track);
    else
      return -1;
}

/*******************************************************************************
    STOP AFTER TRACK
*******************************************************************************/
MEDIA::TrackPtr PlaylistModel::stopAfterTrack() const
{
    return m_stop_after_track;
}

void PlaylistModel::setStopAfterTrack(MEDIA::TrackPtr track)
{
    /* unset old track if exist */
    if(!m_stop_after_track.isNull() && (m_stop_after_track != track)) {
      m_stop_after_track->isStopAfter = !m_stop_after_track->isStopAfter;

      /* emit change */
      int row = m_tracks.indexOf(m_stop_after_track);
      QModelIndex newIndex = index(row, 0, QModelIndex());
      emit dataChanged(newIndex, newIndex);
    }

    /* get new track */
    m_stop_after_track = track;
    m_stop_after_track->isStopAfter = !m_stop_after_track->isStopAfter;


    /* emit change */
    int row = m_tracks.indexOf(m_stop_after_track);
    QModelIndex newIndex = index(row, 0, QModelIndex());
    emit dataChanged(newIndex, newIndex);
}

/*******************************************************************************
    addMediaItem
    addMediaItems
    insertMediaItem
    removeDuplicate
*******************************************************************************/
void PlaylistModel::addMediaItem(const MEDIA::TrackPtr media)
{
    //Debug::debug() << " --- PlaylistModel->addMediaItem";
    m_mutex.lock();
    beginInsertRows(QModelIndex(), m_tracks.size(), m_tracks.size());
    m_tracks << MEDIA::TrackPtr(media);
    endInsertRows();
    m_mutex.unlock();

    if(media->isPlaying)
      m_playing_track = media;
}

void PlaylistModel::addMediaItems(QList<MEDIA::TrackPtr> list)
{
    //Debug::debug() << " --- PlaylistModel->addMediaItems";
    m_mutex.lock();
    beginInsertRows(QModelIndex(), m_tracks.size(), m_tracks.size());

    foreach (const MEDIA::TrackPtr media, list) 
    {
      m_tracks << media;

      if(media->isPlaying)
        m_playing_track = media;
    }

    endInsertRows();
    m_mutex.unlock();
}

void PlaylistModel::insertMediaItem(const MEDIA::TrackPtr media, int pos)
{
    const int start = (pos == -1) ? rowCount() : pos;
    //Debug::debug() << "--- PlaylistModel-->insertMediaItem media pos:" << start;

    m_mutex.lock();
    beginInsertRows(QModelIndex(), start, start);
    m_tracks.insert(start, MEDIA::TrackPtr(media));
    endInsertRows();
    m_mutex.unlock();

    if(media->isPlaying)
      m_playing_track = media;
}

void PlaylistModel::removeDuplicate()
{
    //Debug::warning() << "--- PlaylistModel-->removeDuplicate";
    QSet<QUrl> filenames;

    foreach(MEDIA::TrackPtr media, m_tracks) {
        QString path = media->url;
        if(filenames.contains( path ))
        {
          int row = m_tracks.indexOf(media);
          removeRows(row,1,QModelIndex());
        }
        else
        {
          filenames.insert( path );
        }
    }
}

void PlaylistModel::signalUpdate()
{
    //Debug::warning() << "--- PlaylistModel-->signalUpdate";
    // signal update est utilisé dans le cas d'un changement de mode d'affichage
    // layoutAboutToBeChanged est obligatoire vis à vis du proxy model sinon crash
    emit layoutAboutToBeChanged();
    emit layoutChanged();
    emit updated();
}

/*******************************************************************************
    PlaylistModel::getListDuration
*******************************************************************************/
QString PlaylistModel::getListDuration()
{
    //Debug::debug() << " --- PlaylistModel->getListDuration";

    int seconds = 0;
    foreach (MEDIA::TrackPtr track, m_tracks) {
      if(track->type() != TYPE_TRACK)
        continue;

      if(track->duration > 0)
        seconds += track->duration;
    }
    return UTIL::durationToString(seconds);
}


/*******************************************************************************
    DRAG AND DROP
    -> supportedDropActions
    -> flags
    -> mimeTypes
    -> mimeData
    -> dropMimeData
*******************************************************************************/
Qt::DropActions PlaylistModel::supportedDropActions() const
{
    return Qt::MoveAction | Qt::CopyAction | Qt::LinkAction;
}


Qt::ItemFlags PlaylistModel::flags(const QModelIndex &index) const
{
  Qt::ItemFlags flags = Qt::ItemIsEnabled | Qt::ItemIsSelectable;

  if (index.isValid())
    return flags | Qt::ItemIsDragEnabled;

  return Qt::ItemIsDropEnabled;
}

QStringList PlaylistModel::mimeTypes() const
{
    // Encoding Exported Data with mimeTypes
    QStringList types;
    types << MEDIA_MIME;
    types << "text/uri-list"; //we do accept urls
    return types;
}

QMimeData* PlaylistModel::mimeData( const QModelIndexList &indexes ) const
{
    MediaMimeData* mime = new MediaMimeData();
    mime->setSource(SOURCE_PLAYQUEUE);

    foreach( const QModelIndex &index, indexes ) {
        int row = index.row();

        if (row >= 0 && row < m_tracks.size())
            mime->addTrack( m_tracks.at(row) );
    }

    return mime;
}

//!------------PlaylistModel::dropMimeData -------------------------------------
// --> Inserting Dropped Data into a Model
bool PlaylistModel::dropMimeData(const QMimeData *data,
                                 Qt::DropAction action, int row, int column,
                                 const QModelIndex &parent)
{
    if (action == Qt::IgnoreAction) return true;

    if (data->hasFormat(MEDIA_MIME)) {
      //Debug::debug() << "--- PlaylistModel::dropMimeData --> data->hasFormat(MEDIA_MIME)";
      const MediaMimeData* mediaMimeData = dynamic_cast<const MediaMimeData*>( data );
      if(!mediaMimeData)
        return false;

      // internal move
      if(mediaMimeData->_source == SOURCE_PLAYQUEUE) {
        move(mediaMimeData, row);
      }
      // media item from collection
      else {
        m_task_manager->playlistAddMediaItems(mediaMimeData->getTracks(), row);
      }
      return true;
    }
    else if (data->hasUrls()) {
      m_task_manager->playlistAddUrls(data->urls(), row);
      return true;
    }

    return false;
}

//!------------PlaylistModel::move ---------------------------------------------
void PlaylistModel::move(const MediaMimeData *data, int pos)
{
    Debug::debug() << "PlaylistModel::move";
    emit layoutAboutToBeChanged();

    QList<MEDIA::TrackPtr> moved_items = data->getTracks();

    SortByRowAsc sorter(this);
    qSort(moved_items.begin(), moved_items.end(),  sorter);

    if (pos < 0)
      pos = moved_items.count();

    int start  = pos;
    foreach (MEDIA::TrackPtr track, moved_items) {
      int source_row = rowForTrack(track);
      m_tracks.takeAt(source_row);
      if (pos > source_row)
        start --;
      
      m_tracks.insert(start, track);
      start++;
    }

    emit layoutChanged();
    emit needSelectionAfterMove(moved_items);
}



/*******************************************************************************
    PlaylistModel::slot_sort
*******************************************************************************/
void PlaylistModel::slot_sort(QVariant query)
{
    Debug::debug() << "PlaylistModel::slot_sort";
    emit layoutAboutToBeChanged();
    
    MediaSearch search = MediaSearch::fromQVariant(query);

    SearchEngine* search_engine = new SearchEngine();
    search_engine->init_search_engine(search);
    search_engine->doSearch(true);

    QList<MEDIA::TrackPtr> tracks = search_engine->result();

    m_tracks.clear();
    m_tracks << tracks;
    
    emit layoutChanged();
}


/*
********************************************************************************
*                                                                              *
*    Class PlaylistQueue                                                       *
*                                                                              *
********************************************************************************
*/

PlaylistModel* Playqueue::INSTANCE = 0;

Playqueue::Playqueue()
{
      INSTANCE = new PlaylistModel();
}


