/****************************************************************************************
*  YAROCK                                                                               *
*  Copyright (c) 2010-2013 Sebastien amardeilh <sebastien.amardeilh+yarock@gmail.com>   *
*                                                                                       *
*  This program is free software; you can redistribute it and/or modify it under        *
*  the terms of the GNU General Public License as published by the Free Software        *
*  Foundation; either version 2 of the License, or (at your option) any later           *
*  version.                                                                             *
*                                                                                       *
*  This program is distributed in the hope that it will be useful, but WITHOUT ANY      *
*  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A      *
*  PARTICULAR PURPOSE. See the GNU General Public License for more details.             *
*                                                                                       *
*  You should have received a copy of the GNU General Public License along with         *
*  this program.  If not, see <http://www.gnu.org/licenses/>.                           *
*****************************************************************************************/

#include "StreamView/stream_scene.h"
#include "StreamView/stream_editor.h"
#include "StreamView/stream_loader.h"

#include "StreamModel/streammodel.h"
#include "StreamModel/service_icecast.h"
#include "StreamModel/service_shoutcast.h"
#include "StreamModel/service_tunein.h"
#include "StreamModel/service_xspf.h"

#include "CollectionView/virtualplayqueue.h"
#include "CollectionView/graphicitem.h"

#include "Playlist/playlistmodel.h"
#include "Playlist/task_manager.h"

#include "MediaItem/mediaitem.h"
#include "MediaItem/playlist_parser.h"
#include "networkaccess.h"
#include "utilities.h"

#include "button_item.h"

#include "Widget/statuswidget.h"

#include "global_actions.h"
#include "yarocksettings.h"
#include "debug.h"

/*
********************************************************************************
*                                                                              *
*    Class StreamScene                                                         *
*                                                                              *
********************************************************************************
*/
StreamScene::StreamScene(QWidget *parent)
{
    /* scene configuration */
    _parentWidget = parent;
    this->setParent(_parentWidget);
    this->setItemIndexMethod(QGraphicsScene::NoIndex);
    this->setStickyFocus(true);

    /* init model */
    m_model = StreamModel::instance();
    connect(m_model, SIGNAL(signalPlayingStatusChanged()), this, SLOT(update()));

    /* init player */
    _player = EnginePlayer::instance();

    /* init stream services */
    m_services.insert(MAIN::ViewIceCast,      new IceCast() );
    m_services.insert(MAIN::ViewShoutCast,    new ShoutCast() );
    m_services.insert(MAIN::ViewTuneIn,       new TuneIn() );
    m_services.insert(MAIN::ViewCustomRadio,  new XspfStreams() );
    m_services.insert(MAIN::ViewFavoriteRadio,  m_services.value(MAIN::ViewCustomRadio) );

    foreach (Service* service, m_services.values() ) {
      connect(service, SIGNAL(stateChanged()), this, SIGNAL(repopulateStreamScene()));      
    }    
   
    /*  graphic item context menu */
    m_graphic_item_menu = new GraphicsItemMenu(0);
    connect(m_graphic_item_menu, SIGNAL(menu_action_triggered(ENUM_ACTION_ITEM_MENU)), this, SLOT(slot_contextmenu_triggered(ENUM_ACTION_ITEM_MENU)), Qt::DirectConnection);

    /*  ressources init */
    m_view_mode           = MAIN::ViewIceCast;
    m_mouseGrabbedItem    = 0;
    m_infoSize            = 0;

    /*  scene actions */
    m_actions.insert("duplicate", new QAction(QIcon(),QString(tr("Show duplicate")),this));
    m_actions.insert("reload", new QAction(QIcon(":/images/rebuild.png"),QString("reload"),this));
    m_actions.insert("edit", new QAction(QIcon(":/images/edit_info.png"),QString("edit"),this));
    m_actions.value("duplicate")->setCheckable(true);
    m_actions.value("duplicate")->setChecked(false);
    m_model->setDuplicate(false);
    connect(m_actions.value("duplicate"), SIGNAL(triggered()), this, SLOT(slot_remove_duplicate_stream()));
    connect(m_actions.value("edit"), SIGNAL(triggered()), this, SLOT(slot_edit_stream_list()));
    connect(m_actions.value("reload"), SIGNAL(triggered()), this, SLOT(slot_reload_stream_list()));

    //! drag items (action defined in collection scene)
    connect(ACTIONS()->value(BROWSER_ITEM_MOUSE_MOVE), SIGNAL(triggered()), this, SLOT(slot_item_mouseMove()), Qt::DirectConnection);
}

StreamScene::~StreamScene()
{
    clear();
}



/*******************************************************************************
    playStream
*******************************************************************************/
void StreamScene::playStream()
{
    if(!m_mouseGrabbedItem) return;
    StreamGraphicItem *item = static_cast<StreamGraphicItem*>(m_mouseGrabbedItem);

    const QString streamUrl = item->media->url;

    /*--------------------------------------------------*/
    /* urls simple                                      */
    /* -------------------------------------------------*/
    if( !MEDIA::isPlaylistFile(streamUrl) &&
        !MEDIA::isShoutCastUrl(streamUrl) &&
        !MEDIA::isTuneInUrl(streamUrl) )
    {
      Debug::debug() << "StreamScene -> playStream  we can play it's an url"<< streamUrl;

      VirtualPlayqueue::instance()->appendMedia(item->media);
      return;
    }
    
    /*--------------------------------------------------*/
    /* playlist remote or shoutcast                     */
    /* -------------------------------------------------*/
    Debug::debug() << "StreamScene -> playStream  it's a playlist -> download to do" ;

    uint i = StatusWidget::instance()->startProgressMessage(tr("Loading stream"));
    m_messageIds.insert("StreamLoad", i);
    
    StreamLoader* streamloader = new StreamLoader(item->media);
    streamloader->start_asynchronous_download(item->media->url);
    connect(streamloader, SIGNAL(download_done(MEDIA::TrackPtr)), this, SLOT(slot_streams_fetched(MEDIA::TrackPtr)));
}

void StreamScene::slot_streams_fetched(MEDIA::TrackPtr track)
{
    QList<MEDIA::TrackPtr> tracks;
    foreach(MEDIA::MediaPtr media, track->children()) {
      tracks << MEDIA::TrackPtr::staticCast(media);
      //Debug::debug() << "StreamScene::slot_streams_fetched URL=" << MEDIA::TrackPtr::staticCast(media)->url;
    }
    
    if(!tracks.isEmpty())
      VirtualPlayqueue::instance()->appendMedias(tracks,0);
    
    StreamLoader* loader = qobject_cast<StreamLoader*>(sender());
    if(loader) delete loader;
    
    if (m_messageIds.contains("StreamLoad"))
      StatusWidget::instance()->stopProgressMessage( m_messageIds.take("StreamLoad") );    
}

/*******************************************************************************
    playSelected
*******************************************************************************/
void StreamScene::playSelected()
{
    //Debug::debug() << "StreamScene::playSelected";
    QList<MEDIA::TrackPtr> tracks;

    QList<QGraphicsItem*> list = sortedSelectedItem();
    
    if(list.isEmpty())
      return;
          
    QGraphicsItem* gi = list.first();

    if (gi->type() != GraphicsItem::StreamType) 
      return;
    
    StreamGraphicItem *item = static_cast<StreamGraphicItem*>(gi);
    QString streamUrl = item->media->url;

    /*--------------------------------------------------*/
    /* urls simple                                      */
    /* -------------------------------------------------*/
    if( !MEDIA::isPlaylistFile(streamUrl) &&
        !MEDIA::isShoutCastUrl(streamUrl) &&
        !MEDIA::isTuneInUrl(streamUrl) )
    {
        VirtualPlayqueue::instance()->appendMedia(item->media);
        return;
    }
      
    /*--------------------------------------------------*/
    /* playlist remote or shoutcast                     */
    /* -------------------------------------------------*/
    Debug::debug() << "StreamScene -> playStream  it's a playlist -> download to do" ;

    uint i = StatusWidget::instance()->startProgressMessage(tr("Loading stream"));
    m_messageIds.insert("StreamLoad", i);
    
    StreamLoader* streamloader = new StreamLoader(item->media);
    streamloader->start_asynchronous_download(item->media->url);
    connect(streamloader, SIGNAL(download_done(MEDIA::TrackPtr)), this, SLOT(slot_streams_fetched(MEDIA::TrackPtr)));
}



/*******************************************************************************
    enqueueSelected
*******************************************************************************/
void StreamScene::enqueueSelected(bool is_new_playqueue)
{
    //Debug::debug() << "StreamScene::enqueueSelected";
    QList<MEDIA::TrackPtr> tracks;

    foreach(QGraphicsItem* gi, sortedSelectedItem())
    {
      if (gi->type() == GraphicsItem::StreamType)
      {
        StreamGraphicItem *item = static_cast<StreamGraphicItem*>(gi);
        tracks << item->media;
      }
    } // end foreach

    if(!tracks.isEmpty()) {
      if(is_new_playqueue)
        Playqueue::instance()->clear();
    
      Playqueue::instance()->manager()->playlistAddMediaItems(tracks);
    }
}


/*******************************************************************************
    enqueueStream
*******************************************************************************/
void StreamScene::enqueueStream(bool is_new_playqueue)
{
    if(!m_mouseGrabbedItem) return;

    if(is_new_playqueue)
      Playqueue::instance()->clear();

    StreamGraphicItem *item = static_cast<StreamGraphicItem*>(m_mouseGrabbedItem);

    const QString streamUrl = item->media->url;

    /*--------------------------------------------------*/
    /* urls simple                                      */
    /* -------------------------------------------------*/
    if( !MEDIA::isPlaylistFile(streamUrl) &&
        !MEDIA::isShoutCastUrl(streamUrl) &&
        !MEDIA::isTuneInUrl(streamUrl) )
    {
      Playqueue::instance()->addMediaItem(item->media);
      return;
    }

    /*--------------------------------------------------*/
    /* playlist remote or shoutcast                     */
    /* -------------------------------------------------*/
    // use playlist task manager to load streams into playqueue    
    Playqueue::instance()->manager()->playlistAddFile(streamUrl);
}

/*******************************************************************************
    startStreamsDrag
*******************************************************************************/
void StreamScene::startStreamsDrag(QGraphicsItem* i)
{
    if(!i) // multiple tracks drags
    {
      MediaMimeData* mimedata = new MediaMimeData();

      foreach(QGraphicsItem* gi, sortedSelectedItem())
      {
        StreamGraphicItem *item = static_cast<StreamGraphicItem *>(gi);
        mimedata->addTrack(item->media);
      }

      QDrag *drag = new QDrag(_parentWidget);
      drag->setMimeData(mimedata);
      drag->exec();
    }
    else // single drag
    {
        StreamGraphicItem *item = static_cast<StreamGraphicItem *>(i);
        item->startDrag(_parentWidget);
    }
}

/*******************************************************************************
    sortedSelectedItem
*******************************************************************************/
QList<QGraphicsItem*> StreamScene::sortedSelectedItem()
{
    QList<QGraphicsItem*> selected = selectedItems();
    int count = selected.count();

    bool sort_ok = false;

    do
    {
      sort_ok = false;
      for(int i =0; i<=count-2; i++)
      {
        QGraphicsItem* gi1 = selected.at(i);
        QGraphicsItem* gi2 = selected.at(i+1);

        QPoint p1 = gi1->scenePos().toPoint();
        QPoint p2 = gi2->scenePos().toPoint();

        bool lessthan;
        if(p1.y() < p2.y())
          lessthan = true;
        else if (p1.y() > p2.y())
          lessthan = false;
        else if (p1.x() < p2.x())
          lessthan = true;
        else
          lessthan = false;

        if(!lessthan) {
          selected.swap(i,i+1);
          sort_ok = true;
        }
      } // end for
    } while(sort_ok);

    return selected;
}


/*******************************************************************************
    populateScene
*******************************************************************************/
void StreamScene::populateScene()
{
    Debug::debug() << "StreamScene -> PopulateScene";

    /* clear the scene FIRST */
    clear();
    
    /* update actions */
    m_actions.value("edit")->setEnabled( m_services[m_view_mode]->type() == SERVICE::LOCAL);    
    m_actions.value("reload")->setEnabled( m_services[m_view_mode]->type() != SERVICE::LOCAL);    
    
    /* update local file reader service */
    if(m_services[m_view_mode]->type() == SERVICE::LOCAL) 
    {
      QString path;
      if(m_view_mode == MAIN::ViewCustomRadio)
        path = QString(UTIL::CONFIGDIR + "/radio/" + SETTINGS()->_radiofile);
      else if (m_view_mode == MAIN::ViewFavoriteRadio)
        path = QString(UTIL::CONFIGDIR + "/radio/#favorite.xspf");

      XspfStreams* xspf = static_cast<XspfStreams*>(m_services[MAIN::ViewCustomRadio]);    
      if(xspf->filename() != path)
        xspf->set_filename(path);
    }

    /* handle service state */
    SERVICE::State state = m_services[m_view_mode]->state();
        
    if(state == SERVICE::DATA_OK) 
    {
      m_model->setStreams( m_services[m_view_mode]->streams() );
      if(m_services[m_view_mode]->hasLinks())
        populateExtendedStreamScene();
      else
        populateStreamScene();
    }
    else if (state == SERVICE::NO_DATA || state == SERVICE::DOWNLOADING) 
    {
      populateWaitingScene(
        QString(QObject::tr("downloading streams from %1....please wait...")).arg(m_services[m_view_mode]->name()) 
        );
      m_services[m_view_mode]->load();
    }
    else if (state == SERVICE::ERROR) 
    {
      populateWaitingScene(
        QString(QObject::tr("error accessing stream from %1 !")).arg(m_services[m_view_mode]->name())
        );
    }
     
    /* update playing item */     
    if(_player->engineState() == ENGINE::PLAYING)
        m_model->updatePlayingItem(_player->playingTrack());

    /*  we need to ajust SceneRect */
    int diff = _parentWidget->height() - itemsBoundingRect().height() + 40;
    int ajust  = ( diff > 0) ? diff : 0;
    setSceneRect ( itemsBoundingRect().adjusted(-10, -10, +10, ajust-10) );
}


/*******************************************************************************
    populateStreamScene
*******************************************************************************/
void StreamScene::populateStreamScene()
{
    Debug::debug() << "StreamScene -> populateStreamScene";

    int streamRow      = 0;
    int categoriesRow  = 0;
    m_infoSize         = 0;

    QGraphicsView* view = qobject_cast<QGraphicsView*> (_parentWidget);
    int utilWidth = view->verticalScrollBar()->isVisible() ? (view->width() - view->verticalScrollBar()->width())  : view->width();
    
    /*  header item */
    HeaderItem* header = new HeaderItem();
    header->_width = utilWidth - 10;
    header->setText( m_services[m_view_mode]->name() );
    header->setPos(0, 5);
   
    if(m_view_mode == MAIN::ViewCustomRadio) {
      XspfStreams* xspf = static_cast<XspfStreams*>(m_services[MAIN::ViewCustomRadio]);    
      header->setText( m_services[m_view_mode]->name() + ": " + QFileInfo(xspf->filename()).baseName() );
    }         
    else if(m_view_mode == MAIN::ViewFavoriteRadio) {
      header->setText( tr("Favorites streams") );
    }
        
    addItem(header);
    
    /*  menu button */
    QMenu* menu = new QMenu();
    menu->addAction(m_actions.value("reload"));
    menu->addAction(m_actions.value("edit"));
    menu->addAction(m_actions.value("duplicate"));

    ButtonMenuItem* ui_menu_button = new ButtonMenuItem();
    ui_menu_button->setPixmap(QPixmap(":/images/settings-white-48x48.png"));
    ui_menu_button->setToolTip(tr("view actions"));
    ui_menu_button->setMenu(menu);
    ui_menu_button->setPos(utilWidth-10-ui_menu_button->width(), 5);
    addItem(ui_menu_button);

    categoriesRow++;

    QString currentCategoryName;
    //! stream model loop
    for ( int i = 0; i < m_model->itemCount(); i++ )
    {
      if(!m_model->isStreamFiltered(i)) continue;

      if(currentCategoryName != m_model->streamAt(i)->categorie)
      {
        //! new category
        currentCategoryName = m_model->streamAt(i)->categorie;

        // hack pour une présentation correcte on rajoute une fausse ligne de stream avant une nouvelle categorie
        streamRow++;
        CategorieGraphicItem *category = new CategorieGraphicItem();
        category->_catName = currentCategoryName;
        category->_width   = _parentWidget->width()-20;
        category->setPos(0 ,categoriesRow*40 + streamRow*20 );

        addItem(category);
        categoriesRow++;
      }

      StreamGraphicItem *stream_item = new StreamGraphicItem();
      stream_item->media = m_model->streamAt(i);
      stream_item->media->isFavorite = (m_view_mode == MAIN::ViewFavoriteRadio);
      stream_item->setPos(30, categoriesRow*40 + streamRow*20);
      stream_item->_width   = _parentWidget->width()-30-20;
      stream_item->setToolTip(stream_item->media->url);

      addItem(stream_item);

      streamRow++;
      m_infoSize++;
    }

    if(m_infoSize==0) {
      streamRow++;

      InfoGraphicItem *info = new InfoGraphicItem();
      info->_text = tr("No entry found");
      info->_width   = _parentWidget->width();
      info->setPos( 0 , 10 + categoriesRow*40 + streamRow*20);
      addItem(info);
    }
}

/*******************************************************************************
    populateExtendedStreamScene
*******************************************************************************/
void StreamScene::populateExtendedStreamScene()
{
    int streamRow      = 0;
    int categoriesRow  = 0;
    m_infoSize         = 0;
    int Xpos = 20,Ypos = 5;

    QGraphicsView* view = qobject_cast<QGraphicsView*> (_parentWidget);
    int utilWidth = view->verticalScrollBar()->isVisible() ? (view->width() - view->verticalScrollBar()->width())  : view->width();

    
    /*  header item */
    HeaderItem* header = new HeaderItem();
    header->_width = utilWidth - 10;
    header->setText( m_services[m_view_mode]->name() );
    header->setPos(0, Ypos);
   
    addItem(header);
    
    /*  menu button */
    QMenu* menu = new QMenu();
    menu->addAction(m_actions.value("reload"));
    menu->addAction(m_actions.value("edit"));
    menu->addAction(m_actions.value("duplicate"));

    ButtonMenuItem* ui_menu_button = new ButtonMenuItem();
    ui_menu_button->setPixmap(QPixmap(":/images/settings-white-48x48.png"));
    ui_menu_button->setToolTip(tr("view actions"));
    ui_menu_button->setMenu(menu);
    ui_menu_button->setPos(utilWidth-10-ui_menu_button->width(), 5);
    addItem(ui_menu_button);

    Ypos += 40;
   
    QString currentCategoryName;
    
    
    QList<MEDIA::LinkPtr> links;
    if(m_view_mode == MAIN::ViewTuneIn)
      links = static_cast<TuneIn*>(m_services[MAIN::ViewTuneIn])->links();    
    else
      links = static_cast<ShoutCast*>(m_services[MAIN::ViewShoutCast])->links();

    /*--------------------------------------------------*/
    /* link                                             */
    /* -------------------------------------------------*/         
    foreach(MEDIA::LinkPtr link, links) 
    {
      if(currentCategoryName != link->categorie)
      {
        if(Xpos > 20) {
          Ypos += 30;
          Xpos = 20;
        }

        //! new category
        currentCategoryName = link->categorie;

        // hack pour une présentation correcte on rajoute une fausse ligne de stream avant une nouvelle categorie
        streamRow++;
        CategorieGraphicItem *category = new CategorieGraphicItem();
        category->_catName = link->categorie;
        category->_width   = _parentWidget->width()-20;
        category->setPos(0 , Ypos + categoriesRow*40);

        addItem(category);
        categoriesRow++;
      }

      ButtonItem* button = new ButtonItem();
      button->setText(link->name);
      QVariant v;
      v.setValue(static_cast<MEDIA::LinkPtr>(link));
      button->setData(v);
    
      if(m_view_mode == MAIN::ViewTuneIn)
        connect(button, SIGNAL(clicked()), static_cast<TuneIn*>(m_services[MAIN::ViewTuneIn]), SLOT(slot_activate_link()));
      else
        connect(button, SIGNAL(clicked()), static_cast<ShoutCast*>(m_services[MAIN::ViewShoutCast]), SLOT(slot_activate_link()));

      if(Xpos + button->width() > _parentWidget->width()-20) {
        Xpos = 20;
        Ypos = Ypos + 30;
      }

      button->setPos(Xpos ,Ypos + categoriesRow*40);
      Xpos = Xpos + button->width() + 20;

      addItem(button);
    } // end foreach link
     
    /*--------------------------------------------------*/
    /* stream list                                      */
    /* -------------------------------------------------*/
    //! stream model loop
    for ( int i = 0; i < m_model->itemCount(); i++ )
    {
      if(!m_model->isStreamFiltered(i)) continue;

      if(currentCategoryName != m_model->streamAt(i)->categorie)
      {
        //! new category
        currentCategoryName = m_model->streamAt(i)->categorie;

        // hack pour une présentation correcte on rajoute une fausse ligne de stream avant une nouvelle categorie
        streamRow++;
        CategorieGraphicItem *category = new CategorieGraphicItem();
        category->_catName = currentCategoryName;
        category->_width   = _parentWidget->width()-20;
        category->setPos(0 ,Ypos + categoriesRow*40 + streamRow*20 );

        addItem(category);
        categoriesRow++;
      }

      StreamGraphicItem *stream_item = new StreamGraphicItem();
      stream_item->media = m_model->streamAt(i);
      stream_item->media->isFavorite = (m_view_mode == MAIN::ViewFavoriteRadio);
      stream_item->setPos(30, Ypos + categoriesRow*40 + streamRow*20);
      stream_item->_width   = _parentWidget->width()-30-20;
      stream_item->setToolTip(stream_item->media->url);

      addItem(stream_item);

      streamRow++;
      m_infoSize++;
    }
}



void StreamScene::populateWaitingScene(QString text)
{
    m_actions.value("edit")->setEnabled(false);
    m_actions.value("duplicate")->setEnabled(false);
      
    QGraphicsView* view = qobject_cast<QGraphicsView*> (_parentWidget);
    int utilWidth = view->verticalScrollBar()->isVisible() ? (view->width() - view->verticalScrollBar()->width())  : view->width();
    
    /*  header item */
    HeaderItem* header = new HeaderItem();
    header->_width = utilWidth - 10;
    header->setText( m_services[m_view_mode]->name() );
    header->setPos(0, 5);
   
    if(m_view_mode == MAIN::ViewCustomRadio) {
      XspfStreams* xspf = static_cast<XspfStreams*>(m_services[MAIN::ViewCustomRadio]);    
      header->setText( m_services[m_view_mode]->name() + ": " + QFileInfo(xspf->filename()).baseName() );
    }         
        
    addItem(header);
          
    /*  menu button */
    QMenu* menu = new QMenu();
    menu->addAction(m_actions.value("reload"));
    menu->addAction(m_actions.value("edit"));
    menu->addAction(m_actions.value("duplicate"));

    ButtonMenuItem* ui_menu_button = new ButtonMenuItem();
    ui_menu_button->setPixmap(QPixmap(":/images/settings-white-48x48.png"));
    ui_menu_button->setToolTip(tr("view actions"));
    ui_menu_button->setMenu(menu);
    ui_menu_button->setPos(utilWidth-10-ui_menu_button->width(), 5);
    addItem(ui_menu_button);


    if(m_services[m_view_mode]->state() == SERVICE::DOWNLOADING) { 
      LoadingGraphicItem *info = new LoadingGraphicItem();
      info->_text = text;
      info->_width   = _parentWidget->width();
      info->setPos( 0 , 30 + 40);
      addItem(info);
    } 
    else {
      InfoGraphicItem *info = new InfoGraphicItem();
      info->_text = text;
      info->_width   = _parentWidget->width();
      info->setPos( 0 , 30 + 40);
      addItem(info);
    }
}


/*******************************************************************************
 Mouse and menu event
*******************************************************************************/
void StreamScene::contextMenuEvent(QGraphicsSceneContextMenuEvent *event)
{
    //Debug::debug() << "StreamScene::contextMenuEvent";
    m_mouseGrabbedItem = this->itemAt(event->scenePos());

    if(!m_mouseGrabbedItem) {
      m_graphic_item_menu->hide();
      event->accept();
      return;
    }

    if(this->selectedItems().isEmpty() || !selectedItems().contains(m_mouseGrabbedItem))
    {
      m_graphic_item_menu->setSelectedItems(QList<QGraphicsItem*>());
      m_graphic_item_menu->setGraphicItem(m_mouseGrabbedItem);
    }
    else
    {
      m_graphic_item_menu->setGraphicItem(0);
      m_graphic_item_menu->setSelectedItems(selectedItems());
    }

    m_graphic_item_menu->updateMenu();
    m_graphic_item_menu->exec(event->screenPos());

    // no need to propagate to item
    event->accept();
}

void StreamScene::slot_contextmenu_triggered(ENUM_ACTION_ITEM_MENU id)
{
    switch(id)
    {
      case STREAM_PLAY         : playStream();            break;
      case STREAM_EDIT         : break;
      case STREAM_QUEUE_END    : enqueueStream(false);    break;
      case STREAM_QUEUE_NEW    : enqueueStream(true);     break;
      case STREAM_FAVORITE     : updateStreamFavorite();  break;

      case SELECTION_PLAY      : playSelected();break;
      case SELECTION_QUEUE_END : enqueueSelected(false);break;
      case SELECTION_QUEUE_NEW : enqueueSelected(true);break;
      default:break;
    }
}

void StreamScene::slot_item_mouseMove()
{
    //Debug::debug() << "StreamScene::slot_item_mouseMove";

    // single drag and drop
    QGraphicsItem *gItem = qvariant_cast<QGraphicsItem *>( (ACTIONS()->value(BROWSER_ITEM_MOUSE_MOVE))->data() );

    if(this->selectedItems().isEmpty() || !selectedItems().contains(gItem))
    {
      switch(gItem->type()) {
        case GraphicsItem::StreamType      : startStreamsDrag(gItem);break;
      }
    }
    // multiple drag and drop
    else
    {
      switch(gItem->type()) {
        case GraphicsItem::StreamType     : startStreamsDrag();break;
      }
    }
}

void StreamScene::mouseDoubleClickEvent ( QGraphicsSceneMouseEvent * event )
{
    //Debug::debug() << "StreamScene::mouseDoubleClickEvent";
    m_mouseGrabbedItem = this->itemAt(event->scenePos());

    if(!m_mouseGrabbedItem || m_mouseGrabbedItem->type() != GraphicsItem::StreamType) {
      event->ignore();
      return;
    }

    if(Qt::ControlModifier == QApplication::keyboardModifiers())
    {
      enqueueStream(false);
    }
    else
    {
      playStream();
    }
    // no need to propagate to item
    event->accept();
}


/*******************************************************************************
    Manage favorite
*******************************************************************************/
void StreamScene::updateStreamFavorite()
{
    //Debug::debug() << "StreamScene -> updateStreamFavorite";

    if(!m_mouseGrabbedItem) return;
    StreamGraphicItem *item = static_cast<StreamGraphicItem*>(m_mouseGrabbedItem);

    bool isFavorite   = item->media->isFavorite;

    
    XspfStreams xspf_favorite(UTIL::CONFIGDIR + "/radio/#favorite.xspf");
    xspf_favorite.load();

    if(isFavorite)  // REMOVE IT
    {
      bool removed = false;
      QList<MEDIA::TrackPtr> currentList = xspf_favorite.streams();

      // remove stream if exist
      foreach (MEDIA::TrackPtr mi, currentList) {
        if( mi->url == item->media->url ) {
          currentList.removeAt(currentList.indexOf(mi));
          removed = true;
          break;
        }
      }

      if(removed) {
        qSort(currentList.begin(), currentList.end(), MEDIA::compareStreamCategorie);

        xspf_favorite.setStreams(currentList);
        xspf_favorite.save();

        populateScene();
      }
    }
    else // ADD IT
    {
      bool exist = false;
      // get current favorite
      QList<MEDIA::TrackPtr> currentList = xspf_favorite.streams();

      // add stream if does not exist
      foreach (MEDIA::TrackPtr mi, currentList) {
        if(mi->url == item->media->url) {
          exist = true;
          break;
        }
      }

      if(!exist) {
        currentList.append( item->media );
        qSort(currentList.begin(), currentList.end(), MEDIA::compareStreamCategorie);

        xspf_favorite.setStreams(currentList);
        xspf_favorite.save();

        populateScene();
      }
    }
}



//! -------------------- slot_reload_stream_list -------------------------------
void StreamScene::slot_reload_stream_list()
{
    if( m_view_mode == MAIN::ViewIceCast   || 
        m_view_mode == MAIN::ViewShoutCast ||
        m_view_mode == MAIN::ViewTuneIn)
     {
       m_services[m_view_mode]->reload();        
     }
    
    populateScene();
}

//! --------------------- slot_remove_duplicate_stream -------------------------
void StreamScene::slot_remove_duplicate_stream()
{
    Debug::debug() << "StreamScene::slot_remove_duplicate_stream";

    bool isChecked = m_actions.value("duplicate")->isChecked();

    m_model->setDuplicate(isChecked);
    
    populateScene();
}



//! --------------------- slot_edit_stream_list --------------------------------
void StreamScene::slot_edit_stream_list()
{
    QString file_path;
    bool isFavorite = false;

    if(m_view_mode == MAIN::ViewCustomRadio)
    {
      file_path = UTIL::CONFIGDIR + "/radio/" + SETTINGS()->_radiofile;
    }
    else  if(m_view_mode == MAIN::ViewFavoriteRadio)
    {
      file_path  = UTIL::CONFIGDIR + "/radio/#favorite.xspf";
      isFavorite = true;
    }
    Debug::debug() << "StreamScene::slot_edit_stream_list -> file_path : " << file_path;

    StreamEditor* stream_editor = new StreamEditor(0);
    connect(stream_editor, SIGNAL(stream_list_saved()), this, SLOT(populateScene()));
    stream_editor->load_stream_file(file_path, isFavorite);

    stream_editor->show();
}

